/*
io_sd_card.c

Hardware routines for interfacing with SD cards on a Ninjapass X9

Copyright (C) 2007 by CJ Bell
siegebell at gmail dot com


This software is provided 'as-is', without any express or implied warranty.  In
no event will the author be held liable for any damages arising from the use of
this software.  Permission is granted to anyone to use this software for any
purpose, including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

 1. The origin of this software must not be misrepresented; you must not claim
    that you wrote the original software. If you use this software in a
    product, an acknowledgment in the product documentation would be
    appreciated but is not required.

 2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

 3. This notice may not be removed or altered from any source distribution.
*/

#include <nds.h>
//#include <stdio.h>
#include "io_x9_card.h"
#include "io_sd_card.h"

const int DATA_TIMEOUT = 10000;
//extern void DumpMemory(unsigned char* data, int bytes);

//#define DO_DEBUG(statements) do { statements; } while(0)               
#define DO_DEBUG(statements)

int Choose()
{
    return 1;
    //do
    //{
    //    scanKeys();        
    //    swiWaitForVBlank();

    //    if(keysDown() & KEY_A)
    //        return 1;
    //    else if(keysDown() & KEY_B)
    //        return 2;
    //} while(true);
}



void SDSendCommand(SDCommand sdCmd, uint32 arg, uint8* response)
{
    // Each byte in var represent one bit -- bit 0 of var[x]. Thus, each group
    // of eight bytes in var is "effectively" one byte from the SD card
    // 
    uint8 var[0x200];
    // Send the SD command (0x60)
    // 0x80 words = 0x200 bytes
    X9CardIO(0x60, sdCmd, arg, 0xA1586000, (uint32*)var, 0x80);

    // Skips the first 7 effective bytes (k = 0x30)
    // 0x30 is the size in bits of an SD-mode response
    // the rest must be resulting data...
    int effectiveBit = 0x30;

    // Search for the first "1"/high bit
    // (This is used to find the first byte, which must start with a "1")
    int start;
    int end;
    while(true)
    {
        DO_DEBUG(if(effectiveBit >= 0x200) iprintf("sd-cmd: eb-out-of-bounds\n"));

        if(!(var[effectiveBit] & 1))
        {// start-bit
            // Start at the byte found
            start = effectiveBit;
            // Use 256 bytes of var to construct 32 effective bytes
            end = effectiveBit + 0x100;
            break;
        }

        ++effectiveBit;

        if(effectiveBit >= 0x200)
        {// Hit the end of the search without success 
            // Skip the start-bit
            start = 1; //(var[effectiveBit-1] ^ 1) & 1; <-- known because of previous if-statement
            // Use 255 bytes of var to construct 31+7/8 effective bytes
            end = 0x100; // effectiveBit - 0x100 --> 0x200 - 0x100 --> 0x100
            break;
        }
    }

    DO_DEBUG(if(end > 0x200) iprintf("sd-cmd: end-out-of-bounds\n"));

    // Covert 256 bytes of var into 32 bytes using the low-bits
    // This doesn't care what sd_cs_data is initially because each entry will
    // be shifted 8 bits.
    if(start < end)
    {
        const uint8* end_ptr = &(var[end]);
        uint8* ptr = &(var[start]);

        int bit = 7;
        int j = 0;
        while(ptr < end_ptr)
        {
            if(bit < 0)
            {// Go to the next effective byte
                bit = 7;
                ++j;
            }

            DO_DEBUG(if(j >= 0x20) iprintf("sd-cmd: j-out-of-bounds\n"));

            response[j] = (response[j] << 1) | (*ptr & 1);

            --bit;
            ++ptr;
        } 

    }
}

bool SDWriteSingleBlock(uint32 address, const void* buffer)
{
    uint32 temp[0x80];

    // Tell the X9 the sd-address which we'll be writing to
    cardIO_4(address);

    // Preemptively send the data before issuing the command so
    // that the X9 can compute the CRC
    X9CardWriteData(0, 0, buffer);

    uint8 response[0x20];
    SDSendCommand(WriteSingleBlock,0,response);

    X9CardIO(0x63, 0, 0, 0xB15863FA, temp, 0x80);

    // TODO: Timeout
    int idx;
    do
    {
        // Repeat SD command?
        X9CardIO(0x62, 0, 0, 0xA1586000, temp, 0x80);

        for(idx = 0; idx < 0x200 && (temp[idx] & 0x10); ++idx)
            ;

    } while(idx == 0x200);


    // TODO: TIMEOUT
    do
    {
        // Repeat SD command?
        X9CardIO(0x62, 0, 0, 0xA1586000, temp, 0x80);

        for(idx = 0; idx < 0x200 && !(temp[idx] & 0x10); ++idx)
            ;

    } while(idx == 0x200);

    return true;

}


bool SDReadSingleBlock(uint32 address, void* destination)
{
    uint8 buffer[0x1000];// split in two halves

    DO_DEBUG(iprintf("CardIO ReadSingleBlock: "));
    // 0x60: send an SD command
    // 0xA3586000 means we don't expect the standard response? (but data) -- serial vs 4-bit? length?
    // 0x200 words = 0x800 bytes
    X9CardIO(0x60, (uint8)ReadSingleBlock, address, 0xA3586000, (uint32*)buffer, 0x200);
    DO_DEBUG(iprintf("done\n"));

    // Search for the start of data
    unsigned int k = 0;
    for(k = 0; k < 0x800 && (buffer[k] & 0xf0); ++k)
        ;

    unsigned int  c = 0;
    unsigned int attempts = 0;
    DO_DEBUG(iprintf("Receive data: "));
    while(k == 0x800)
    {
        if(attempts++ >= DATA_TIMEOUT)
        {
            DO_DEBUG(iprintf("data: giving up"));
            return false;
        }

        // 0x62 means "repeat" last command?
        X9CardIO(0x62, 0, address, 0xA3586000, (uint32*)buffer, 0x200);
        
        c+= 0x800;

        // Keep searching for the response-data
        for(k = 0; k < 0x800 && (buffer[k] & 0xf0); ++k)
            ;
    }
    DO_DEBUG(iprintf("done\n"));

    // Read another chunk because the full 512 bytes haven't yet been received
    if(0x800 - k < 0x400) // k > 0x400    
        X9CardIO(0x62, 0, address, 0xA3586000, (uint32*)&buffer[0x800], 0x200);

    c = k + 1; // just after the 0
    k+= 0x400; // read 0x400 half-bytes, or 0x200 bytes
    ++k;

    // IO is done with half-bytes: reconstruct full bytes
    uint8* dest = (uint8*)destination;
    for(; c < k; c+= 2)
    {
        DO_DEBUG(if(c >= 0x1000) iprintf("sb-read: c-out-of-bounds\n"));
        DO_DEBUG(if(dest >= ((uint8*)destination)+0x200) iprintf("sb-read: data-out-of-bounds\n"));

        *dest++ = (buffer[c] & 0xF0) | (buffer[c+1] >> 4);
    }

    // Why is the last command repeated here?
    // So we can get the response?
    // NOTE: SDSendCommand uses 0xA1586000 and gets a response
    // the data-transfer commands use 0xA3586000 and expect no response...
    X9CardIO(0x62, 0, address, 0xA1586000, (uint32*)buffer, 0x80);

    //DO_DEBUG(DumpMemory(destination, 0x200));

    return true;
}

//bool SDInitialize(unsigned int timeout, uint32* relativeCardAddress)
//{
//    if(!relativeCardAddress)
//        return false;
//
//    int attempts;
//    uint8 response[0x20];
//
//    DO_DEBUG(iprintf("CMD0: "));
//    // CMD0 initialize (reset)
//    SDSendCommand(SetIdle,0,response);
//    DO_DEBUG(iprintf("done\n"));
//
//    // TODO: this should return the correct voltage: use that
//    DO_DEBUG(iprintf("Set"));
//    SDSendCommand(AppCommand,0,response);
//    DO_DEBUG(iprintf("Voltage: "));
//    SDSendAppCommand(ACMD41,0x00200000,response); // 3.3-3.4 VDD Voltage Window
//
//    //ACMD41 - identify, wait til ready..
//    // Should be repeated N times until success, else fail
//    // set voltage?
//    // sd_cs_data set by sendSDCommand()
//    for(attempts = 0; response[1] != 0x80 /* Ready */; ++attempts)
//    {
//        DO_DEBUG(iprintf("nope\n"));
//
//        if(attempts >= timeout)
//        {
//            DO_DEBUG(iprintf("SetVoltage: giving up\n"));
//            return false;
//        }
//
//        DO_DEBUG(iprintf("Set"));
//        SDSendCommand(AppCommand,0,response);
//        DO_DEBUG(iprintf("Voltage: "));
//        SDSendAppCommand(ACMD41,0x00100000,response); // 3.2-3.3 VDD Voltage Window
//    }
//    DO_DEBUG(iprintf("done\n"));
//
//    // Retrieves Card's Name
//    // CMD2 get CID
//    DO_DEBUG(iprintf("GetCID: "));
//    SDSendCommand(GetCID,0,response);
//    DO_DEBUG(iprintf("done\n"));
//
//    // Should be repeated N time for success, else fail
//    attempts = 0;
//    do
//    {
//        DO_DEBUG(if(attempts > 0) iprintf("nope\n"));
//
//        if(attempts >= timeout)
//        {
//            DO_DEBUG(iprintf("GetRCA: giving up\n"));
//            return false;
//        }
//
//        DO_DEBUG(iprintf("GetRCA: "));
//        SDSendCommand(GetRCA,0,response);
//        ++attempts;
//    } while((response[3] & 0x1e) == (SDStandbyState << 1)); // busy
//    DO_DEBUG(iprintf("done\n"));
//
//    *relativeCardAddress = (response[1] << 24) | (response[2] << 16);
//
//
//    DO_DEBUG(iprintf("Get CSD? (A=yes,B=no)\n"));
//    if(Choose() == 1)
//    {
//        // Some cards won't go to higher speeds unless they think you
//        // checked their capabilities
//        DO_DEBUG(iprintf("GetCSD "));
//        SDSendCommand(GetCSD,*relativeCardAddress,response);
//        DO_DEBUG(iprintf("done\n"));
//    }
//
//    //AL go SD data mode
//    DO_DEBUG(iprintf("SelectCard: "));
//    SDSendCommand(SelectCard,*relativeCardAddress,response);
//    DO_DEBUG(iprintf("done\n"));
//
//
//    //AL set SD bus width
//    DO_DEBUG(iprintf("Set"));
//    SDSendCommand(AppCommand,*relativeCardAddress,response);
//    DO_DEBUG(iprintf("BusWidth: "));
//    SDSendAppCommand(SetBusWidth,2,response); // 4-bit bus
//    DO_DEBUG(iprintf("done\n"));
//
//    DO_DEBUG(iprintf("Set BlockLength? (A=yes,B=no)\n"));
//    if(Choose() == 1)
//    {
//        // Use 512 byte blocks
//        DO_DEBUG(iprintf("SetBlockLength: "));
//        SDSendCommand(SetBlockLength,512,response);
//        DO_DEBUG(iprintf("done\n"));
//    }
//
//    return true;
//}
