/*
io_x9sd.c

Hardware routines for interfacing with SD cards on a Ninjapass X9

Copyright (C) 2007 by CJ Bell
siegebell at gmail dot com


This software is provided 'as-is', without any express or implied warranty.  In
no event will the author be held liable for any damages arising from the use of
this software.  Permission is granted to anyone to use this software for any
purpose, including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

 1. The origin of this software must not be misrepresented; you must not claim
    that you wrote the original software. If you use this software in a
    product, an acknowledgment in the product documentation would be
    appreciated but is not required.

 2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

 3. This notice may not be removed or altered from any source distribution.

*/

#include "disc_io.h"
#include "io_x9sd.h"

#ifdef NDS

//#include <stdio.h>
#include <string.h>
#include <nds.h>
#include "io_sd_card.h"
#include "io_x9_card.h"
#include "io_sd_common.h"


bool SDSendCommand6(uint8* response, uint8 command, uint32 data)
{
    uint8 sdResponse[32];
    SDSendCommand((SDCommand)command | 0x40, data, sdResponse);
    memcpy(response, sdResponse, 6);
    return true;
}

bool SDSendCommand17(uint8* response, uint8 command, uint32 data)
{
    uint8 sdResponse[32];
    SDSendCommand((SDCommand)command | 0x40, data, sdResponse);
    memcpy(response, sdResponse, 17);
    return true;
}



static uint32 relativeCardAddress;

//#define DO_DEBUG(statements) do { statements; } while(0)               
#define DO_DEBUG(statements)

//#define ENABLE_WRITE

#ifdef ENABLE_WRITE
// THIS IS A HACK!
// After performing a write, reading from the SD card fails
// Not sure if this is due to an SD-error or in the way I'm
// communicating with the X9
bool sdWriteLast;
#endif


bool _X9SD_isInserted(void)
{
    DO_DEBUG(iprintf("IsInserted()\n"));
    uint8 response[8];
    SDSendCommand(GetStatus, 0, response);

    // Make sure the card responded correctly
    return response[0] == GetStatus;
}

bool _X9SD_clearStatus(void)
{
    DO_DEBUG(iprintf("ClearStatus()\n"));
    //return SDInitialize(5, &relativeCardAddress);
    return _SD_InitCard(&SDSendCommand6, SDSendCommand17, true, &relativeCardAddress);
}

bool _X9SD_shutdown(void)
{
    DO_DEBUG(iprintf("Shutdown()\n"));
    return false;
}

bool _X9SD_startup(void)
{
    DO_DEBUG(iprintf("Startup()\n"));

#ifdef ENABLE_WRITE
    sdWriteLast = false;
#endif
    cardIO_1();
    cardIO_2(0x1B00);
    cardIO_3(0);
    cardIO_4(0);

    uint8 response[0x20];
    SDSendCommand(SetIdle,0,response);

    //return SDInitialize(10, &relativeCardAddress);
    return _SD_InitCard(SDSendCommand6, SDSendCommand17, true, &relativeCardAddress);
} 


bool _X9SD_writeSectors (uint32 sector, uint32 sectorCount, const void* buffer)
{
    DO_DEBUG(iprintf("WriteSectors(%i,%i)\n",sector,sectorCount));

    const uint32 blockLength = 512;
    while(sectorCount > 0)
    {
        SDWriteSingleBlock(sector*blockLength, buffer);
 
        buffer+= blockLength;
        --sectorCount;
        ++sector;
    }
#ifdef ENABLE_WRITE
    sdWriteLast = true; // HACK-- set so that the SD card can be reset on the next read
#endif

    return true;
}

bool _X9SD_readSectors (uint32 sector, uint32 sectorCount, void* buffer)
{    
    DO_DEBUG(iprintf("ReadSectors(%i,%i)\n",sector,sectorCount));

#ifdef ENABLE_WRITE
    if(sdWriteLast)
    {// THIS IS A HACK!
        // After performing a write, reading from the SD card fails
        // Not sure if this is due to an SD-error or in the way I'm
        // communicating with the X9
        uint8 response[32];
        SDSendCommand(AbortReadWrite, 0, response);
        SDSendCommand(SetIdle, 0, response);

        if(!_X9SD_startup())
            return false;

        sdWriteLast = false;
    }
#endif

    const uint32 blockLength = 512;
	while(sectorCount > 0)
	{
        SDReadSingleBlock(sector*blockLength, buffer);
        buffer+= blockLength;
        --sectorCount;
        ++sector;
	}

    return true;

}



IO_INTERFACE io_x9sd = {
	DEVICE_TYPE_X9SD,
	FEATURE_MEDIUM_CANREAD | FEATURE_MEDIUM_CANWRITE | FEATURE_SLOT_NDS,
	(FN_MEDIUM_STARTUP)&_X9SD_startup,
	(FN_MEDIUM_ISINSERTED)&_X9SD_isInserted,
	(FN_MEDIUM_READSECTORS)&_X9SD_readSectors,
	(FN_MEDIUM_WRITESECTORS)&_X9SD_writeSectors,
	(FN_MEDIUM_CLEARSTATUS)&_X9SD_clearStatus,
	(FN_MEDIUM_SHUTDOWN)&_X9SD_shutdown
};

LPIO_INTERFACE X9SD_GetInterface(void) {
	return &io_x9sd;
};




#endif // defined NDS
